theory HelperLemmasAutoCorres
imports
  "autocorres-0.9-beta3/autocorres/AutoCorres"
begin


(* Definitions and lemmas from assignement 3 *)
 

;install_C_file "exam.c"
;autocorres [ts_rules = nondet] "exam.c"

abbreviation
  valid_node :: "new_globals \<Rightarrow> node_C ptr \<Rightarrow> bool"
where
  "valid_node s p \<equiv> \<exists>x. heap_node_C s p = Some x"

abbreviation
  node_next :: "new_globals \<Rightarrow> node_C ptr \<Rightarrow> node_C ptr"
where
  "node_next s p \<equiv> next_C (the (heap_node_C s p))"

abbreviation
  node_value :: "new_globals \<Rightarrow> node_C ptr \<Rightarrow> word32"
where
  "node_value s p \<equiv> value_C (the (heap_node_C s p))"

  

primrec
  linked_list :: "new_globals \<Rightarrow> node_C ptr list \<Rightarrow> node_C ptr \<Rightarrow> bool"
where
    "linked_list s [] p =
			(p = NULL)"
  | "linked_list s (x#xs) p =
			(p \<noteq> NULL \<and> p = x \<and> valid_node s p \<and> linked_list s xs (node_next s p))"

lemma linked_list_NULL [simp]:
    "linked_list s xs NULL = (xs = [])"
  by (case_tac xs, auto)

lemma linked_list_equal_list[iff]:
    "\<lbrakk> linked_list s xs p; linked_list s ys p \<rbrakk> \<Longrightarrow> xs = ys"
  apply (induct xs arbitrary: p ys)
   apply simp
  apply (case_tac ys, auto)
  done
  
lemma linked_list_equal_ptr[iff]:
    "\<lbrakk> linked_list s xs p; linked_list s xs q \<rbrakk> \<Longrightarrow> p = q"
  apply (induct xs arbitrary: p q)
   apply clarsimp
  apply clarsimp
  done
  
lemma linked_list_tail:
    "linked_list s (xs @ ys) p \<Longrightarrow> \<exists>q. linked_list s ys q"
  apply (induct xs arbitrary: p)
   apply force
  apply clarsimp
  done

lemma linked_list_hd_not_in_list:
     "\<lbrakk> linked_list s xs q; node_next s p = q \<rbrakk> \<Longrightarrow> p \<notin> set xs"
  apply (rule ccontr)
  apply clarsimp
  apply (drule split_list)
  apply clarsimp
  apply (frule linked_list_tail)
  apply clarsimp
  apply (drule (1) linked_list_equal_list)
  apply clarsimp
  done

lemma linked_list_distinct:
    "\<lbrakk> linked_list s xs p \<rbrakk> \<Longrightarrow> distinct xs"
  apply (induct xs arbitrary: p)
   apply clarsimp
  apply clarsimp
  apply (auto dest!: linked_list_hd_not_in_list)
  done


definition
  is_linked_list :: "new_globals \<Rightarrow> node_C ptr \<Rightarrow> bool"
where
  "is_linked_list s p \<equiv> (\<exists>xs. linked_list s xs p)"

definition
  the_linked_list :: "new_globals \<Rightarrow> node_C ptr \<Rightarrow> node_C ptr list"
where
  "the_linked_list s p \<equiv> (THE xs. linked_list s xs p)"

lemma is_linked_list:
    "\<lbrakk> linked_list s xs p \<rbrakk> \<Longrightarrow> is_linked_list s p"
  by (fastforce simp: is_linked_list_def)

lemma the_linked_list_val:
    "\<lbrakk> linked_list s xs p \<rbrakk> \<Longrightarrow> the_linked_list s p = xs"
  by (auto simp: the_linked_list_def intro: linked_list_equal_list)


definition
  values_of :: "new_globals \<Rightarrow> node_C ptr list \<Rightarrow> word32 list"
where
  "values_of s xs = map (node_value s) xs"

definition
  list_values :: "new_globals \<Rightarrow> node_C ptr \<Rightarrow> word32 list"
where
  "list_values s p = map (node_value s) (the_linked_list s p)"

lemma the_linked_list_next:
  "\<lbrakk> is_linked_list s next; next \<noteq> NULL \<rbrakk> \<Longrightarrow>
      the_linked_list s next = next # (the_linked_list s (node_next s next))"
  apply (clarsimp simp: is_linked_list_def)
  apply (subst the_linked_list_val, assumption)
  apply (metis linked_list.simps(2) linked_list_NULL
      linked_list_equal_ptr the_linked_list_val not_NilE)
  done

lemma is_linked_list_next [simp]:
  "\<lbrakk> is_linked_list s next; next \<noteq> NULL \<rbrakk> \<Longrightarrow>
      is_linked_list s (node_next s next)"
  apply (clarsimp simp: is_linked_list_def)
  apply (case_tac xs, auto)
  done


lemma the_linked_list_NULL [simp]:
   "the_linked_list s NULL = []"
  by (clarsimp simp: the_linked_list_def)

lemma is_linked_list_NULL [simp]:
   "is_linked_list s NULL"
  by (clarsimp simp: is_linked_list_def)
  
end
