(*
 * Copyright (C) 2013, National ICT Australia Limited. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 *  * The name of National ICT Australia Limited nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *)

theory HeapLift
imports
  TypHeapSimple
  ValidNoFail
  CorresXF
  L2Defs
  ExecConcrete
begin

definition "L2Tcorres st A C = corresXF st (\<lambda>r _. r) (\<lambda>r _. r) \<top> A C"

lemma L2Tcorres_id:
  "L2Tcorres id C C"
  by (metis L2Tcorres_def corresXF_id)

lemma L2Tcorres_fail:
  "L2Tcorres st L2_fail X"
  apply (clarsimp simp: L2Tcorres_def L2_defs)
  apply (rule corresXF_fail)
  done

definition "abs_expr st P A C \<equiv> \<forall>s. P (st s) \<longrightarrow> (C s) = A (st s) "
definition "abs_modifies st P A C \<equiv> \<forall>s. P (st s) \<longrightarrow> st (C s) = A (st s) "

lemma abs_expr_fun_app [heap_abs_fo]:
  "\<lbrakk> abs_expr st Y b' b; abs_expr st X a' a \<rbrakk> \<Longrightarrow>
      abs_expr st (X and Y) (\<lambda>s. a' s (b' s)) (\<lambda>s. ((a s) $ (b s)))"
  apply (clarsimp simp: abs_expr_def)
  done

lemma abs_expr_constant [heap_abs]:
  "abs_expr st \<top> (\<lambda>s. a) (\<lambda>s. a)"
  apply (clarsimp simp: abs_expr_def)
  done

lemma L2Tcorres_modify [heap_abs]:
    "\<lbrakk> abs_modifies st P a c \<rbrakk> \<Longrightarrow> L2Tcorres st (L2_seq (L2_guard P) (\<lambda>_. (L2_modify a))) (L2_modify c)"
  apply (monad_eq simp: L2_defs abs_modifies_def L2Tcorres_def corresXF_def)
  done

lemma L2Tcorres_gets [heap_abs]:
    "abs_expr st P a c \<Longrightarrow> L2Tcorres st (L2_seq (L2_guard P) (\<lambda>_. L2_gets a n)) (L2_gets c n)"
  apply (monad_eq simp: corresXF_def L2Tcorres_def L2_defs abs_expr_def)
  done

lemma L2Tcorres_gets_const [heap_abs]:
    "L2Tcorres st (L2_gets (\<lambda>s. a) n) (L2_gets (\<lambda>s. a) n)"
  apply (monad_eq simp: corresXF_def L2Tcorres_def L2_defs abs_expr_def)
  done

lemma L2Tcorres_guard [heap_abs]:
    "\<lbrakk> abs_expr st P a c \<rbrakk> \<Longrightarrow> L2Tcorres st (L2_guard (P and a)) (L2_guard c)"
  apply (monad_eq simp: corresXF_def L2Tcorres_def L2_defs abs_expr_def)
  done

lemma L2Tcorres_recguard [heap_abs]:
    "\<lbrakk> L2Tcorres st A C \<rbrakk> \<Longrightarrow> L2Tcorres st (L2_recguard n A) (L2_recguard n C)"
  apply (monad_eq simp: corresXF_def L2Tcorres_def L2_defs Ball_def Bex_def split: sum.splits)
  done

lemma L2Tcorres_while [heap_abs]:
  assumes body_corres:  "\<And>x. L2Tcorres st (B' x) (B x)"
  and guard_impl_cond:  "\<And>r. abs_expr st (G r) (C' r) (C r)"
  shows "L2Tcorres st (L2_guarded_while G C' B' i n) (L2_while C B i n)"
proof -
  have cond_match: "\<And>r s. G r (st s) \<Longrightarrow> C' r (st s) = C r s"
    using guard_impl_cond
    by (clarsimp simp: abs_expr_def)

  have "corresXF st (\<lambda>r _. r) (\<lambda>r _. r) (\<lambda>_. True)
           (doE _ \<leftarrow> guardE (G i);
                     whileLoopE C'
                       (\<lambda>i. doE r \<leftarrow> B' i;
                                _ \<leftarrow> guardE (G r);
                                returnOk r
                       odE) i
           odE)
     (whileLoopE C B i)"
    apply (rule corresXF_guard_imp)
     apply (rule corresXF_guarded_while [where P="\<lambda>_ _. True" and P'="\<lambda>_ _. True"])
         apply (clarsimp cong: corresXF_cong)
         apply (rule corresXF_guard_imp)
          apply (rule body_corres [unfolded L2Tcorres_def])
         apply simp
        apply (clarsimp simp: cond_match)
       apply clarsimp
       apply (rule hoareE_TrueI)
      apply simp
     apply simp
    apply simp
    done

  thus ?thesis
    by (clarsimp simp: L2Tcorres_def L2_defs
            guardE_def returnOk_liftE)
qed

definition "abs_spec st P (A :: ('a \<times> 'a) set) (C :: ('c \<times> 'c) set)
           \<equiv> (\<forall>s t. P (st s) \<longrightarrow> (((s, t) \<in> C) \<longrightarrow> ((st s, st t) \<in> A)))
                              \<and> (\<forall>s. P (st s) \<longrightarrow> (\<exists>x. (st s, x) \<in> A) \<longrightarrow> (\<exists>x. (s, x) \<in> C))"

lemma L2Tcorres_spec [heap_abs]:
  "\<lbrakk> abs_spec st P A C \<rbrakk>
     \<Longrightarrow> L2Tcorres st (L2_seq (L2_guard P) (\<lambda>_. (L2_spec A))) (L2_spec C)"
  apply (monad_eq simp: corresXF_def L2Tcorres_def L2_defs image_def set_eq_UNIV
             split_def Ball_def state_select_def abs_spec_def split: sum.splits)
  done

lemma abs_spec_constant [heap_abs]:
  "abs_spec st \<top> {(a, b). C} {(a, b). C}"
  apply (clarsimp simp: abs_spec_def)
  done

lemma L2Tcorres_condition [heap_abs]:
  "\<lbrakk> L2Tcorres st L L';
     L2Tcorres st R R';
     abs_expr st P C' C \<rbrakk> \<Longrightarrow>
   L2Tcorres st (L2_seq (L2_guard P) (\<lambda>_. L2_condition C' L R)) (L2_condition C L' R')"
  apply (clarsimp simp: L2_defs abs_expr_def L2Tcorres_def)
  apply (rule corresXF_exec_abs_guard [unfolded guardE_def])
  apply (rule corresXF_cond)
    apply (metis corresXF_guard_imp)
   apply (metis corresXF_guard_imp)
  apply simp
  done

lemma L2Tcorres_seq [heap_abs]:
  "\<lbrakk> L2Tcorres st L' L; \<And>r. L2Tcorres st (\<lambda>s. R' r s) (\<lambda>s. R r s) \<rbrakk>
      \<Longrightarrow> L2Tcorres st (L2_seq L' (\<lambda>r s. R' r s)) (L2_seq L (\<lambda>r s. R r s))"
  apply (clarsimp simp: L2Tcorres_def L2_defs)
  apply (rule corresXF_guard_imp)
  apply (erule corresXF_join [where P'="\<lambda>x y s. x = y" and Q="\<lambda>_. True"])
     apply (metis (full_types) corresXF_assume_pre)
    apply simp
    apply (rule hoareE_TrueI)
   apply simp
  apply simp
  done

lemma L2Tcorres_catch [heap_abs]:
    "\<lbrakk> L2Tcorres st L L';
      \<And>r. L2Tcorres st (\<lambda>s. R r s) (\<lambda>s. R' r s)
     \<rbrakk> \<Longrightarrow> L2Tcorres st (L2_catch L (\<lambda>r s. R r s)) (L2_catch L' (\<lambda>r s. R' r s))"
  apply (clarsimp simp: L2Tcorres_def L2_defs)
  apply (rule corresXF_guard_imp)
  apply (erule corresXF_except [where P'="\<lambda>x y s. x = y" and Q="\<lambda>_. True"])
     apply (metis (full_types) corresXF_assume_pre)
    apply simp
    apply (rule hoareE_TrueI)
   apply simp
  apply simp
  done

lemma L2Tcorres_unknown [heap_abs]:
  "L2Tcorres st (L2_unknown name) (L2_unknown name)"
  apply (clarsimp simp: L2_unknown_def selectE_def[symmetric])
  apply (clarsimp simp: L2Tcorres_def)
  apply (auto intro!: corresXF_select_select)
  done

lemma L2Tcorres_throw [heap_abs]:
  "L2Tcorres st (L2_throw x n) (L2_throw x n)"
  apply (clarsimp simp: L2Tcorres_def L2_defs)
  apply (rule corresXF_throw)
  apply simp
  done

lemma L2Tcorres_split [heap_abs]:
  "\<lbrakk> \<And>x y. L2Tcorres st (P x y) (P' x y) \<rbrakk> \<Longrightarrow>
    L2Tcorres st (case a of (x, y) \<Rightarrow> P x y) (case a of (x, y) \<Rightarrow> P' x y)"
  apply (clarsimp simp: split_def)
  done

lemma L2Tcorres_seq_unused_result [heap_abs]:
  "\<lbrakk> L2Tcorres st L L'; L2Tcorres st R R' \<rbrakk> \<Longrightarrow> L2Tcorres st (L2_seq L (\<lambda>_. R)) (L2_seq L' (\<lambda>_. R'))"
  apply (rule L2Tcorres_seq, auto)
  done

lemma abs_expr_split [heap_abs]:
  "\<lbrakk> \<And>a b. abs_expr st (P a b) (A a b) (C a b) \<rbrakk>
       \<Longrightarrow> abs_expr st (case r of (a, b) \<Rightarrow> P a b)  (case r of (a, b) \<Rightarrow> A a b) (case r of (a, b) \<Rightarrow> C a b)"
  apply (auto simp: split_def)
  done

lemma L2Tcorres_recguard_0:
    "L2Tcorres st (L2_recguard 0 A) C"
  apply (monad_eq simp: corresXF_def L2Tcorres_def L2_defs)
  done

lemma L2Tcorres_abstract_fail [heap_abs]:
  "L2Tcorres st L2_fail L2_fail"
  apply (clarsimp simp: L2Tcorres_def L2_defs)
  apply (rule corresXF_fail)
  done

lemma abs_expr_id [heap_abs]:
  "abs_expr id \<top> A A"
  apply (clarsimp simp: abs_expr_def)
  done

lemma abs_modify_id [heap_abs]:
  "abs_modifies id \<top> A A"
  apply (clarsimp simp: abs_modifies_def)
  done

definition "hl_measure_valid st A C \<equiv> \<forall>s. A (st s) = C s"

(* eta-contract LHS of exec_gets *)
lemma exec_gets2: "(gets f >>= m) = (\<lambda>s. m (f s) s)"
  apply (rule ext)
  apply (rule exec_gets)
  done

(* We don't mark some of these lemmas [heap_abs] for use in heap_lift,
   because they unify too easily. We instantiate them manually.
   NB: hl_measure_valid goals must come first; they are instantiated during heap lifting. *)
lemma L2Tcorres_exec_concrete:
  "L2Tcorres id (\<lambda>s. A (Ma s) s) (\<lambda>s. C (Mc s) s) \<Longrightarrow> L2Tcorres st (exec_concrete st (L2_call Ma A)) (L2_call Mc C)"
  apply (clarsimp simp: L2Tcorres_def L2_call_def)
  apply (rule corresXF_exec_concrete)
  apply (rule corresXF_except)
     apply (monad_eq simp: L2_defs liftE_bindE exec_gets2)
     apply assumption
    apply (rule corresXF_fail)
   apply wp[1]
  apply simp
  done

lemmas L2Tcorres_exec_concrete_undefined [heap_abs] =
  L2Tcorres_exec_concrete [where Ma = "\<lambda>_. undefined" and Mc = "\<lambda>_. undefined"]

(* NB: We don't need a recguard_dec instantiation for L2Tcorres_exec_{concrete,abstract},
   because functions in a recursive group are always lifted to the same heap. *)

lemma L2Tcorres_exec_concrete_measured:
  "\<lbrakk> hl_measure_valid id Ma Mc; \<And>m. L2Tcorres id (A m) (C m) \<rbrakk>
    \<Longrightarrow> L2Tcorres st (exec_concrete st (L2_call Ma A)) (L2_call Mc C)"
  apply (unfold hl_measure_valid_def)
  apply (rule L2Tcorres_exec_concrete)
  apply (simp add: L2Tcorres_def corresXF_def)
  done


lemma L2Tcorres_exec_abstract:
    "L2Tcorres st (\<lambda>s. A (Ma s) s) (\<lambda>s. C (Mc s) s) \<Longrightarrow> L2Tcorres id (exec_abstract st (L2_call Ma A)) (L2_call Mc C)"
  apply (clarsimp simp: L2_call_def L2Tcorres_def)
  apply (rule corresXF_exec_abstract)
  apply (rule corresXF_except)
     apply (monad_eq simp: L2_defs liftE_bindE exec_gets2)
     apply assumption
    apply (rule corresXF_fail)
   apply wp[1]
  apply simp
  done

lemmas L2Tcorres_exec_abstract_undefined [heap_abs] =
  L2Tcorres_exec_abstract [where Ma = "\<lambda>_. undefined" and Mc = "\<lambda>_. undefined"]

lemma L2Tcorres_exec_abstract_measured:
  "\<lbrakk> hl_measure_valid st Ma Mc; \<And>m. L2Tcorres st (A m) (C m) \<rbrakk>
    \<Longrightarrow> L2Tcorres id (exec_abstract st (L2_call Ma A)) (L2_call Mc C)"
  apply (unfold hl_measure_valid_def)
  apply (rule L2Tcorres_exec_abstract)
  apply (simp add: L2Tcorres_def corresXF_def)
  done


lemma L2Tcorres_call:
  "\<lbrakk> L2Tcorres st (\<lambda>t. A (Ma t) t) (\<lambda>s. C (Mc s) s) \<rbrakk>
     \<Longrightarrow> L2Tcorres st (L2_call Ma A) (L2_call Mc C)"
  unfolding L2Tcorres_def L2_call_def
  apply (monad_eq simp: L2_defs liftE_bindE exec_gets2)
  apply (rule corresXF_except)
     apply simp
    apply (rule corresXF_fail)
   apply (rule hoareE_TrueI)
  apply simp
  done

lemma L2Tcorres_rec_call:
  "\<lbrakk> L2Tcorres st (A m) (C m) \<rbrakk>
     \<Longrightarrow> L2Tcorres st (L2_call (\<lambda>_. m) A) (L2_call (\<lambda>_. m) C)"
  apply (rule L2Tcorres_call)
  apply fast
  done

lemma L2Tcorres_measured_call:
  "\<lbrakk> hl_measure_valid st Ma Mc; \<And>m. L2Tcorres st (A m) (C m) \<rbrakk>
    \<Longrightarrow> L2Tcorres st (L2_call Ma A) (L2_call Mc C)"
  apply (rule L2Tcorres_call)
  unfolding L2Tcorres_def hl_measure_valid_def
  apply (rule corresXF_I)
    apply (metis corresXF_exec_normal)
   apply (metis corresXF_exec_except)
  apply (metis corresXF_exec_fail)
  done

lemma L2Tcorres_call_undefined [heap_abs]:
  "\<lbrakk> L2Tcorres st (A undefined) (C undefined) \<rbrakk>
    \<Longrightarrow> L2Tcorres st (L2_call (\<lambda>_. undefined) A) (L2_call (\<lambda>_. undefined) C)"
  by (rule L2Tcorres_call, assumption)


lemma abs_expr_lambda_null [heap_abs]:
  "abs_expr st P A C \<Longrightarrow> abs_expr st P (\<lambda>s r. A s) (\<lambda>s r. C s)"
  apply (clarsimp simp: abs_expr_def)
  done

(*
 * Assert the given abstracted heap (accessed using "getter" and "setter") for type
 * "'a" is a valid abstraction w.r.t. the given state translation functions.
 *)

definition
  valid_typ_heap ::
    "('s \<Rightarrow> 't) \<Rightarrow>
       ('t \<Rightarrow> ('a::c_type) ptr \<Rightarrow> 'a option) \<Rightarrow>
       (('a ptr \<Rightarrow> 'a option) \<Rightarrow> 't \<Rightarrow> 't) \<Rightarrow>
       ('s \<Rightarrow> heap_raw_state) \<Rightarrow>
        ((heap_raw_state \<Rightarrow> heap_raw_state) => 's => 's) \<Rightarrow>
       bool"
where
  "valid_typ_heap st getter setter t_hrs t_hrs_update \<equiv>
     (\<forall>s x. getter (setter x s) = x)
     \<and> (\<forall>s. setter (getter s) s = s)
     \<and> (\<forall>s p x. getter (st s) p = Some x \<longrightarrow> c_guard p)
     \<and> (\<forall>s p x. getter (st s) p = Some x \<longrightarrow>
           h_val (hrs_mem (t_hrs s)) p = x)
     \<and> (\<forall>s p v v'. getter (st s) p = Some v \<longrightarrow>
           st (t_hrs_update (hrs_mem_update (heap_update p v')) s) =
                           setter ((getter (st s))(p := Some v')) (st s))"

lemma valid_typ_heapI [intro!]:
  assumes getter_setter_idem: "\<And>s x. getter (setter x s) = x"
  and setter_getter_idem: "\<And>s. setter (getter s) s = s"
  and getter_implies_safe: "\<And>s p x. getter (st s) p = Some x \<Longrightarrow> c_guard p"
  and getter_data_correct: "\<And>s p x. getter (st s) p = Some x \<Longrightarrow> h_val (hrs_mem (t_hrs s)) p = x"
  and abs_update_matches_conc_update:
      "\<And>s p v v' x. getter (st s) p = Some v \<Longrightarrow>
           st (t_hrs_update (hrs_mem_update (heap_update p v')) s) =
                           setter ((getter (st s))(p := Some v')) (st s)"
  shows "valid_typ_heap st getter setter t_hrs t_hrs_update"
  apply (clarsimp simp: valid_typ_heap_def)
  apply (safe | fact | rule ext)+
  done

(*
 * Assert the given field ("field_getter", "field_setter") of the given structure
 * can be abstracted into the heap, and then accessed as a HOL object.
 *)

definition
  valid_struct_field
    :: "('s \<Rightarrow> 't)
           \<Rightarrow> string list
           \<Rightarrow> ('p \<Rightarrow> ('f::c_type))
           \<Rightarrow> ('f \<Rightarrow> 'p \<Rightarrow> 'p)
           \<Rightarrow> ('t \<Rightarrow> (('p::c_type) ptr \<Rightarrow> 'p option))
           \<Rightarrow> ((('p ptr \<Rightarrow> 'p option) \<Rightarrow> ('p ptr \<Rightarrow> 'p option)) \<Rightarrow> 't \<Rightarrow> 't)
           \<Rightarrow> ('s \<Rightarrow> heap_raw_state)
           \<Rightarrow> ((heap_raw_state \<Rightarrow> heap_raw_state) \<Rightarrow> 's \<Rightarrow> 's)
           \<Rightarrow> bool"
where
  "valid_struct_field st field_name field_getter field_setter
            getter setter t_hrs t_hrs_update \<equiv>
      (\<forall>s p. (\<exists>x. getter (st s) p = Some x) \<longrightarrow>
          h_val (hrs_mem (t_hrs s)) (Ptr &(p\<rightarrow>field_name))
              = field_getter (the (getter (st s) p)))
      \<and> (\<forall>s p val. (\<exists>x. getter (st s) p = Some x) \<longrightarrow>
                 st (t_hrs_update (hrs_mem_update (heap_update (Ptr &(p\<rightarrow>field_name)) val)) s) =
                           setter (\<lambda>old. old(p := Some (field_setter val (the (old p))))) (st s))
      \<and> (\<forall>s p x. getter (st s) p = Some x \<longrightarrow> c_guard p)
      \<and> (\<forall>p. c_guard (p :: 'p ptr) \<longrightarrow> c_guard (Ptr &(p\<rightarrow>field_name) :: 'f ptr))"

lemma valid_struct_fieldI [intro]:
  fixes st :: "'s \<Rightarrow> 't"
  fixes field_getter :: "('a::c_type) \<Rightarrow> ('f::c_type)"
  shows "\<lbrakk> \<And>s p x. getter (st s) p = Some x \<Longrightarrow>
        h_val (hrs_mem (t_hrs s)) (Ptr &(p\<rightarrow>field_name))  = field_getter (the (getter (st s) p));
     \<And>s p val x. getter (st s) p = Some x \<Longrightarrow>
        st (t_hrs_update (hrs_mem_update (heap_update (Ptr &(p\<rightarrow>field_name)) val)) s) =
             setter (\<lambda>old. old(p := Some (field_setter val (the (old p))))) (st s);
     \<And>s p x. getter (st s) p = Some x \<Longrightarrow> c_guard p;
     \<And>(p::'a ptr). c_guard p \<Longrightarrow> c_guard (Ptr &(p\<rightarrow>field_name) :: 'f ptr) \<rbrakk> \<Longrightarrow>
    valid_struct_field st field_name field_getter field_setter getter setter t_hrs t_hrs_update"
  apply (fastforce simp: valid_struct_field_def)
  done

lemma valid_typ_heap_get_setD:
  "valid_typ_heap st getter setter t_hrs t_hrs_update \<Longrightarrow>
       getter (setter x s) = x"
  apply (clarsimp simp: valid_typ_heap_def)
  done

lemma valid_typ_heap_set_getD:
  "valid_typ_heap st getter setter t_hrs t_hrs_update \<Longrightarrow>
       setter (getter s) s = s"
  apply (clarsimp simp: valid_typ_heap_def)
  done

lemma valid_typ_heap_get_hvalD:
  "valid_typ_heap st getter setter t_hrs t_hrs_update \<Longrightarrow>
       getter (st s) p = Some x \<Longrightarrow> h_val (hrs_mem (t_hrs s)) p = x"
  apply (clarsimp simp: valid_typ_heap_def)
  done

lemma valid_typ_heap_t_hrs_updateD:
  "valid_typ_heap st getter setter t_hrs t_hrs_update \<Longrightarrow>
       getter (st s) p = Some v \<Longrightarrow>
           st (t_hrs_update (hrs_mem_update (heap_update p v')) s) =
                           setter ((getter (st s))(p := Some v')) (st s)"
  apply (clarsimp simp: valid_typ_heap_def)
  done

lemma heap_abs_expr_guard [heap_abs]:
  "\<lbrakk> valid_typ_heap st getter setter t_hrs t_hrs_update; abs_expr st P x' x \<rbrakk> \<Longrightarrow>
     abs_expr st (P and (\<lambda>s. \<exists>x. getter s (x' s) = Some x)) (\<lambda>s. True) (\<lambda>s. (c_guard (x s :: ('a::{c_type}) ptr)))"
  apply (clarsimp simp: abs_expr_def simple_lift_def heap_ptr_valid_def valid_typ_heap_def)
  apply blast
  done

lemma heap_abs_expr_h_val [heap_abs]:
  "\<lbrakk> valid_typ_heap st getter setter t_hrs t_hrs_update;
     abs_expr st P x' x \<rbrakk> \<Longrightarrow>
      abs_expr st
       (P and (\<lambda>s. \<exists>a. getter s (x' s) = Some a))
         (\<lambda>s. the (getter s (x' s)))
         (\<lambda>s. (h_val (hrs_mem (t_hrs s))) (x s))"
  apply (clarsimp simp: abs_expr_def simple_lift_def)
  apply (metis valid_typ_heap_get_hvalD)
  done

lemma heap_abs_modifies_heap_update [heap_abs]:
  "\<lbrakk>  valid_typ_heap st getter setter t_hrs t_hrs_update;
     abs_expr st Pb b' b;
     abs_expr st Pc c' c \<rbrakk> \<Longrightarrow>
      abs_modifies st (Pb and Pc and (\<lambda>s. \<exists>x. getter s (b' s) = Some x))
        (\<lambda>s. setter ((getter s)(b' s := Some (c' s))) s)
           (\<lambda>s. t_hrs_update (hrs_mem_update (heap_update (b s :: ('a::c_type) ptr) (c s :: 'a))) s)"
  apply (clarsimp simp: typ_simple_heap_simps abs_expr_def abs_modifies_def)
  apply (metis valid_typ_heap_t_hrs_updateD)
  done

lemma abs_expr_field_getter [heap_abs]:
  "\<lbrakk> valid_struct_field st field_name field_getter field_setter
                     getter setter t_hrs t_hrs_setter;
      abs_expr st P a c \<rbrakk> \<Longrightarrow>
   abs_expr st (P and (\<lambda>s. \<exists>x. getter s (a s) = Some x)) (\<lambda>s. field_getter (the (getter s (a s))))
              (\<lambda>s. h_val (hrs_mem (t_hrs s)) (Ptr &((c s)\<rightarrow>field_name)))"
  apply (clarsimp simp: abs_expr_def valid_struct_field_def valid_typ_heap_def)
  done

lemma abs_expr_field_setter [heap_abs]:
  "\<lbrakk> valid_struct_field st field_name
          field_getter field_setter getter setter t_hrs t_hrs_update;
     abs_expr st P p p'; abs_expr st Q val val' \<rbrakk> \<Longrightarrow>
  abs_modifies st (P and Q and (\<lambda>s. \<exists>x. getter s (p s) = Some (x::'a::{c_type})))
      (\<lambda>s. setter (\<lambda>old. old((p s) := Some (field_setter (val s) (the (old (p s)))))) s)
      (\<lambda>s. t_hrs_update (hrs_mem_update (heap_update (Ptr &((p' s)\<rightarrow>field_name)) (val' s))) s)"
  apply (clarsimp simp: abs_expr_def valid_struct_field_def valid_typ_heap_def abs_modifies_def)
  done

lemma abs_expr_field_guard [heap_abs]:
  "\<lbrakk> valid_struct_field st field_name
          (field_getter :: 'p \<Rightarrow> 'f) field_setter getter setter t_hrs t_hrs_update;
     abs_expr st P p p' \<rbrakk> \<Longrightarrow>
  abs_expr st (P and (\<lambda>s. \<exists>x. getter s (p s) = Some (x::'p::{c_type})))
      (\<lambda>s. True)
      (\<lambda>s. c_guard (Ptr &((p' s)\<rightarrow>field_name) :: 'f::{c_type} ptr))"
  apply (clarsimp simp: abs_expr_def)
  apply (clarsimp simp: abs_expr_def valid_struct_field_def valid_typ_heap_def)
  apply (erule allE2, erule impE, force)+
  apply (erule allE, erule impE, force)
  apply clarsimp
  done

(*
 * Convert gets/sets to global variables into gets/sets in the new globals record.
 *)

definition
  valid_globals_field :: "
     ('s \<Rightarrow> 't)
     \<Rightarrow> ('s \<Rightarrow> 'a)
     \<Rightarrow> (('a \<Rightarrow> 'a) \<Rightarrow> 's \<Rightarrow> 's)
     \<Rightarrow> ('t \<Rightarrow> 'a)
     \<Rightarrow> (('a \<Rightarrow> 'a) \<Rightarrow> 't \<Rightarrow> 't)
     \<Rightarrow> bool"
where
  "valid_globals_field st old_getter old_setter new_getter new_setter \<equiv>
    (\<forall>s. new_getter (st s) = old_getter s)
    \<and> (\<forall>s v. new_setter v (st s) = st (old_setter v s))"

lemma abs_expr_globals_getter [heap_abs]:
  "\<lbrakk> valid_globals_field st old_getter old_setter new_getter new_setter \<rbrakk>
    \<Longrightarrow> abs_expr st \<top> new_getter old_getter"
  apply (clarsimp simp: valid_globals_field_def abs_expr_def)
  done

lemma abs_expr_globals_setter [heap_abs]:
  "\<lbrakk> valid_globals_field st old_getter old_setter new_getter new_setter;
     \<And>old. abs_expr st (P old) (v old) (v' old) \<rbrakk>
    \<Longrightarrow> abs_modifies st (\<lambda>s. \<forall>old. P old s) (\<lambda>s. new_setter (\<lambda>old. v old s) s) (\<lambda>s. old_setter (\<lambda>old. v' old s) s)"
  apply (clarsimp simp: valid_globals_field_def abs_expr_def abs_modifies_def)
  done

(* Signed words are stored on the heap as unsigned words. *)

lemma uint_scast [simp]:
    "uint (scast x :: 'a word) = uint (x :: 'a::len signed word)"
  apply (subst down_cast_same [symmetric])
   apply (clarsimp simp: cast_simps)
  apply (subst uint_up_ucast)
   apply (clarsimp simp: cast_simps)
  apply simp
  done

lemma to_bytes_signed_word:
    "to_bytes (x :: 'a::len8 signed word) p = to_bytes (scast x :: 'a word) p"
  by (clarsimp simp: to_bytes_def typ_info_word word_rsplit_def)

lemma from_bytes_signed_word:
    "length p = len_of TYPE('a) div 8 \<Longrightarrow>
           (from_bytes p :: 'a::len8 signed word) = ucast (from_bytes p :: 'a word)"
  by (clarsimp simp: from_bytes_def word_rcat_def
              scast_def cast_simps typ_info_word)

lemma hrs_mem_update_signed_word:
    "hrs_mem_update (heap_update (ptr_coerce p :: 'a::len8 word ptr) (scast val :: 'a::len8 word))
               = hrs_mem_update (heap_update p (val :: 'a::len8 signed word))"
  apply (rule ext)
  apply (clarsimp simp: hrs_mem_update_def split_def)
  apply (clarsimp simp: heap_update_def to_bytes_signed_word
             size_of_def typ_info_word)
  done

lemma h_val_signed_word:
    "(h_val a p :: 'a::len8 signed word) = ucast (h_val a (ptr_coerce p :: 'a word ptr))"
  apply (clarsimp simp: h_val_def)
  apply (subst from_bytes_signed_word)
   apply (clarsimp simp: size_of_def typ_info_word)
  apply (clarsimp simp: size_of_def typ_info_word)
  done

lemma abs_expr_signed_word [heap_abs]:
  "\<lbrakk> valid_typ_heap st getter setter t_hrs t_hrs_update;
     abs_expr st P x' x \<rbrakk> \<Longrightarrow>
      abs_expr st
       (P and (\<lambda>s. \<exists>a. getter s (ptr_coerce (x' s)) = Some a))
         (\<lambda>s. ucast (the (getter s (ptr_coerce (x' s :: 'a signed word ptr) :: 'a word ptr))))
         (\<lambda>s. (h_val (hrs_mem (t_hrs s))) (x s :: 'a::len8 signed word ptr))"
  apply (drule heap_abs_expr_h_val [where P=P
     and x'="\<lambda>s. ptr_coerce (x' s)"
     and x="\<lambda>s. ptr_coerce (x s)"])
   apply (clarsimp simp: abs_expr_def)
  apply (clarsimp simp: abs_expr_def)
  apply (drule spec, erule (1) impE)
  apply (drule spec, erule impE, fastforce)
  apply (clarsimp simp: hrs_mem_def h_val_signed_word split: option.splits)
  done

lemma abs_modifies_signed_word [heap_abs]:
  "\<lbrakk>  valid_typ_heap st getter setter t_hrs t_hrs_update;
     abs_expr st Pb b' b;
     abs_expr st Pc c' c \<rbrakk> \<Longrightarrow>
      abs_modifies st (Pb and Pc and (\<lambda>s. \<exists>x. getter s (ptr_coerce (b' s)) = Some x))
        (\<lambda>s. setter ((getter s)(ptr_coerce  (b' s) := Some (scast (c' s) :: 'a word))) s)
           (\<lambda>s. t_hrs_update (hrs_mem_update (heap_update (b s :: 'a::len8 signed word ptr) (c s :: 'a signed word))) s)"
  apply (drule heap_abs_modifies_heap_update [
      where b'="\<lambda>s. ptr_coerce (b' s)"
      and b="\<lambda>s. ptr_coerce (b s)"
      and c'="\<lambda>s. scast (c' s)"
      and c="\<lambda>s. scast (c s)"
      and Pb=Pb and Pc=Pc])
    apply (clarsimp simp: abs_expr_def)
   apply (clarsimp simp: abs_expr_def)
  apply (clarsimp simp: abs_expr_def abs_modifies_def)
  apply (drule spec, erule (1) impE)
  apply (drule spec, erule (1) impE)
  apply (drule spec, erule impE, fastforce)
  apply (subst (asm) hrs_mem_update_signed_word)
  apply clarsimp
  done

end
