(*
 * Copyright (C) 2013, National ICT Australia Limited. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 *  * The name of National ICT Australia Limited nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *)

(*
 * Final simplifications applied to function specs.
 *)

theory Polish
imports TypeStrengthen
begin

lemma L2polish [polish]:
  "L2_seq = bindE"
  "L2_unknown n = unknownE"
  "L2_gets g n = getsE g"
  "L2_modify = modifyE"
  "L2_condition = condition"
  "L2_catch = handleE'"
  "L2_while C B i n = whileLoopE C B i"
  "L2_spec s = (specE s >>=E (\<lambda>x. unknownE))"
  "L2_fail = fail"
  "L2_throw v n = throwError v"
  apply (auto simp: L2_defs skipE_def modifyE_def unknownE_def
      getsE_def specE_def bind_liftE_distrib selectE_def intro!: ext)
  done

lemma returnOk_skip [polish]: "returnOk () = skipE"
  by (clarsimp simp: skipE_def)

lemma return_skip [polish]: "return () = skip"
  by (clarsimp simp: skip_def)

lemma liftE_foo_to_fooE [polish]:
  "liftE (unknown) = unknownE"
  "liftE (gets x) = getsE x"
  "liftE (return a) = returnOk a"
  "liftE (modify m) = modifyE m"
  "liftE (spec s) = specE s"
  "liftE skip = skipE"
  "liftE (guard g) = guardE g"
  apply (monad_eq simp: unknownE_def selectE_def unknown_def
    getsE_def modifyE_def specE_def skip_def skipE_def guardE_def)+
  done

lemma gets_to_return [polish]:
  "gets (\<lambda>x. a) = return a"
  apply (clarsimp simp: simpler_gets_def return_def)
  done

lemma getsE_to_returnOk [polish]:
  "getsE (\<lambda>x. a) = returnOk a"
  apply (clarsimp simp: getsE_def simpler_gets_def returnOk_def2 return_def)
  done

lemma select_UNIV_unknown [polish]: "select UNIV = unknown"
  by (clarsimp simp: unknown_def)
lemma selectE_UNIV_unknown [polish]: "liftE (select UNIV) = unknownE"
  by (clarsimp simp: unknownE_def selectE_def)

lemma unknown_unit [polish,simp]: "(unknown :: ('s, unit) nondet_monad) = skip"
  apply (rule ext)+
  apply (clarsimp simp: skip_def select_def return_def skipE_def unknown_def)
  apply force
  done

lemma unknownE_unit [polish,simp]: "(unknownE :: ('s, 'a + unit) nondet_monad) = skipE"
  apply (rule ext)+
  apply (monad_eq simp: skip_def selectE_def skipE_def unknownE_def)
  done

lemma condition_to_when [polish]:
  "condition (\<lambda>s. C) A skip = when C A"
  "condition (\<lambda>s. C) A' skipE = whenE C A'"
  apply (auto simp: condition_def when_def skip_def skipE_def whenE_def)
  done

lemma condition_to_unless [polish]:
  "condition (\<lambda>s. C) skip A = unless C A"
  "condition (\<lambda>s. C) skipE A' = unlessE C A'"
  apply (auto simp: condition_def unless_def when_def skip_def skipE_def unlessE_def)
  done

lemma bind_skip [simp, polish]:
  "(x >>= (\<lambda>_. skip)) = x"
  by (monad_eq simp: skip_def)

lemma bindE_skipE [simp, polish]:
  "(x >>=E (\<lambda>_. skipE)) = x"
  by (monad_eq simp: skipE_def)

lemma skip_bind [simp, polish]:
  "(skip >>= P) = (P ())"
  by (monad_eq simp: skip_def)

lemma skipE_bindE [simp, polish]:
  "(skipE >>=E P) = (P ())"
  by (monad_eq simp: skip_def skipE_def)

lemma ogets_to_oreturn [polish]: "ogets (\<lambda>s. P) = oreturn P"
  apply (clarsimp simp: ogets_def oreturn_def)
  done

lemma ocondition_ret_ret [polish]:
  "ocondition P (oreturn A) (oreturn B) = ogets (\<lambda>s. if P s then A else B)"
  by (auto intro!:ext simp: ocondition_def ogets_def)

lemma ocondition_gets_ret [polish]:
  "ocondition P (ogets A) (oreturn B) = ogets (\<lambda>s. if P s then A s else B)"
  by (auto intro!:ext simp: ocondition_def ogets_def)

lemma ocondition_ret_gets [polish]:
  "ocondition P (oreturn A) (ogets B) = ogets (\<lambda>s. if P s then A else B s)"
  by (auto intro!:ext simp: ocondition_def ogets_def)

lemma ocondition_gets_gets [polish]:
  "ocondition P (ogets A) (ogets B) = ogets (\<lambda>s. if P s then A s else B s)"
  by (auto intro!:ext simp: ocondition_def ogets_def)

lemma bindE_K_bind [polish]: "A >>=E (\<lambda>x :: unit. B) = (A >>=E K_bind B)"
  apply (clarsimp simp: K_bind_def)
  done
lemma bind_K_bind [polish]: "A >>= (\<lambda>x :: unit. B) = (A >>= K_bind B)"
  apply (clarsimp simp: K_bind_def)
  done
lemma obind_K_bind [polish]: "A |>> (\<lambda>x :: unit. B) = (A  |>> K_bind B)"
  apply (clarsimp simp: K_bind_def)
  done

lemma condition_to_if [polish]:
  "condition (\<lambda>s. C) (return a) (return b) = return (if C then a else b)"
  apply (clarsimp simp: condition_splits)
  done

declare condition_fail_rhs [polish]
declare condition_fail_lhs [polish]
declare simple_bind_fail [polish]
declare simple_bindE_fail [polish]
declare condition_bind_fail [polish]

lemma simple_K_bind_fail [polish]:
  "(guard X >>= K_bind (fail)) = fail"
  "(modify M >>= K_bind (fail)) = fail"
  "(return X >>= K_bind (fail)) = fail"
  "(gets X >>= K_bind (fail)) = fail"
  "(skip >>= K_bind (fail)) = fail"
  apply (auto simp: skip_def)
  done

lemma simple_K_bindE_fail [polish]:
  "(guardE X >>=E K_bind (fail)) = fail"
  "(modifyE M >>=E K_bind (fail)) = fail"
  "(returnOk X >>=E K_bind (fail)) = fail"
  "(getsE X >>=E K_bind (fail)) = fail"
  "(skipE >>=E K_bind (fail)) = fail"
  apply auto
  done

declare whileLoop_fail [polish]
declare whileLoopE_fail [polish]
declare owhile_fail [polish]

lemma oguard_True [simp, polish]: "oguard (\<lambda>_. True) = oreturn ()"
  by (clarsimp simp: oreturn_def oguard_def)

lemma oguard_False [simp, polish]: "oguard (\<lambda>_. False) = ofail"
  by (clarsimp simp: ofail_def oguard_def)

lemma infinite_option_while': "(Some x, Some y) \<notin> option_while' (\<lambda>_. True) B"
  apply (rule notI)
  apply (induct "Some x :: 'a option" "Some y :: 'a option"
      arbitrary: x y rule: option_while'.induct)
  apply auto
  done

lemma owhile_infinite_loop [polish]:
    "owhile (\<lambda>r s. C) B x = (oguard (\<lambda>s. \<not> C) |>> (\<lambda>_. oreturn x))"
  apply (case_tac C)
   apply (rule ext)
   apply (clarsimp simp: owhile_def option_while_def obind_def split: option.splits)
   apply (metis infinite_option_while' None_not_eq option_while'_THE)
  apply (subst owhile_unroll)
  apply (clarsimp simp: obind_def oreturn_def split: option.splits)
  done

declare obind_return [polish]
declare bind_return [polish]
declare bindE_returnOk [polish]

declare fail_bind [polish]
declare fail_bindE [polish]

lemmas returnOk_bindE_throwError [polish]
  = returnOk_bindE [where f=throwError]

declare singleton_iff [polish]
declare when_True [polish]
declare when_False [polish]
declare unless_True [polish]
declare unless_False [polish]
declare recguard_dec_def [polish]

lemma let_triv [polish]: "(let x = y in x) = y"
  apply (clarsimp simp: Let_def)
  done

lemma ucast_scast_same [polish]: "ucast ((scast x :: ('a::len) word)) = (x :: 'a word)"
  apply (clarsimp simp: ucast_def scast_def)
  done

(* Optimising "if" constructs. *)

lemma return_if_P_1_0_bind [polish]:
    "(return (if P then 1 else 0) >>= (\<lambda>x. Q x))
        = (return P >>= (\<lambda>x. Q (if x then 1 else 0)))"
  apply simp
  done

lemma returnOk_if_P_1_0_bindE [polish]:
    "(returnOk (if P then 1 else 0) >>=E (\<lambda>x. Q x))
        = (returnOk P >>=E (\<lambda>x. Q (if x then 1 else 0)))"
  apply simp
  done

lemma gets_if_P_1_0_bind [polish]:
    "(gets (\<lambda>s. if P s then 1 else 0) >>= (\<lambda>x. Q x))
        = (gets P >>= (\<lambda>x. Q (if x then 1 else 0)))"
  apply (auto intro!: monad_eqI simp: in_gets in_bind snd_gets snd_bind Bex_def)
  done

lemma getsE_if_P_1_0_bindE [polish]:
    "(getsE (\<lambda>s. if P s then 1 else 0) >>=E (\<lambda>x. Q x))
        = (getsE P >>=E (\<lambda>x. Q (if x then 1 else 0)))"
  apply (auto intro!: monad_eqI simp: getsE_def in_gets
        in_liftE in_bindE snd_gets snd_bindE snd_liftE Bex_def)
  done

lemma if_P_1_0_neq_0 [polish, simp]:
    "((if P then 1 else (0::('a::{zero_neq_one}))) \<noteq> 0) = P"
  apply simp
  done

lemma if_P_1_0_eq_0 [polish, simp]:
    "((if P then 1 else (0::('a::{zero_neq_one}))) = 0) = (\<not> P)"
  apply simp
  done

lemma bind_return_bind [polish]:
    "(A >>= (\<lambda>x. (return x >>= (\<lambda>y. B x y)))) = (A >>= (\<lambda>x. B x x))"
  by simp

lemma bindE_returnOk_bindE [polish]:
    "(A >>=E (\<lambda>x. (returnOk x >>=E (\<lambda>y. B x y)))) = (A >>=E (\<lambda>x. B x x))"
  by simp

lemma obind_oreturn_obind [polish]:
    "(A |>> (\<lambda>x. (oreturn x |>> (\<lambda>y. B x y)))) = (A |>> (\<lambda>x. B x x))"
  by simp

declare if_distrib [where f=scast, polish, simp]
declare if_distrib [where f=ucast, polish, simp]
declare if_distrib [where f=unat, polish, simp]
declare if_distrib [where f=uint, polish, simp]
declare if_distrib [where f=sint, polish, simp]

declare cast_simps [polish]

end
